/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#error No gmp.h!
#endif /* HAVE_GMP_H */

#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif

#include <mhash.h>
#include "thread.h"
#include "tiger.h"
#include "sha1.h"


#ifndef _CRYPT_H
#define _CRYPT_H

#define XOR_ARRAY(c,a,b,len)  do { \
	int i; \
	for (i = 0; i < (len); i++) { \
		(c)[i] = (a)[i]^(b)[i]; \
	} \
} while (0)
#define FOLD_SHA1_160(a)	XOR_ARRAY((a),(a),((a)+10),10)
#define FOLD_TIGER_160(a)	XOR_ARRAY((a),(a),((a)+10),10)
#define FOLD_TIGER_192(a)	XOR_ARRAY((a),(a),((a)+12),12)


#define DEFAULT_RANDOM		"/dev/urandom"


#define RIJNDAEL_128	1
#define TWOFISH_128	2

#define RIJNDAEL_128_BLOCK_SIZE	16
#define RIJNDAEL_128_KEY_SIZE	16
#define RIJNDAEL_128_IV_SIZE	16
#define TWOFISH_128_BLOCK_SIZE	16
#define TWOFISH_128_KEY_SIZE	16
#define TWOFISH_128_IV_SIZE	16

typedef struct cipher {
	MCRYPT __td;

	int type;
} cipher_t;


static pthread_once_t	__cipher_init_once;
static object_t		*__cipher_init_mutex;
void __cipher_init(void);


cipher_t *cipher_create(int,void *,void *);
void cipher_free(cipher_t *);
void cipher_init(cipher_t *,int,void *,void *);
void cipher_destroy(cipher_t *);
void cipher_encrypt(cipher_t *,void *,int);
void cipher_decrypt(cipher_t *,void *,int);
#define cipher_get_type(c)		(*c).type
#define cipher_get_key_size(t)		16
#define cipher_get_block_size(t)	16


#define IPAD	0x36
#define	OPAD	0x5C

#define SHA1_160	1
#define TIGER_192	2
#define TIGER_160	3
#define TIGER_128	4

#define SHA1_160_BLOCK_SIZE	64
#define SHA1_160_DIGEST_SIZE	20
#define TIGER_192_BLOCK_SIZE	64
#define TIGER_192_DIGEST_SIZE	24
#define TIGER_160_BLOCK_SIZE	64
#define TIGER_160_DIGEST_SIZE	20
#define TIGER_128_BLOCK_SIZE	64
#define TIGER_128_DIGEST_SIZE	16

#define HASH(type,digest,plaintext,size) do { \
	hash_t h; \
	hash_init((type),&h); \
	hash_update(&h,(plaintext),(size)); \
	hash_final(&h,(digest)); \
} while (0)

#define HMAC(type,digest,key,key_size,plaintext,plaintext_size) do { \
	hash_t h; \
	hmac_init((type),&h,(key),(key_size)); \
	hash_update(&h,(plaintext),(plaintext_size)); \
	hmac_final(&h,(digest)); \
} while (0)

typedef struct hash {
	int type;
	unsigned char ctx[MAX(sizeof(TIGER_CTX),sizeof(SHA_CTX))];
	int block_size;
	int digest_size;

	/* for HMAC */
	unsigned char key[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	unsigned char ipad[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	unsigned char opad[MAX(SHA1_160_BLOCK_SIZE,TIGER_192_BLOCK_SIZE)];
	int key_size;
} hash_t;

void hash_init(int,hash_t *);
void hash_update(hash_t *,void *,int);
void hash_final(hash_t *,void *);

void hmac_init(int,hash_t *,void *,int);
void hmac_update(hash_t *,void *,int);
void hmac_final(hash_t *,void *);

int hash_get_digest_size(int);
int hash_get_block_size(int);


#define MHASH_HMAC(hashid,hmac,hmac_len,secret,secret_len,data,data_len) \
	do {			\
		MHASH td;	\
		void *digest;	\
		td = mhash_hmac_init((hashid),(void *)(secret),(secret_len),mhash_get_hash_pblock(hashid));	\
		mhash(td,(data),(data_len));	\
		digest = mhash_hmac_end(td);	\
		memcpy((void *)(hmac),digest,(size_t)(hmac_len));	\
		free(digest);	\
	} while(0)

#define MHASH_HMAC_QUICK(hashid,hmac,secret,secret_len,data,data_len) \
	do {			\
		MHASH td;	\
		td = mhash_hmac_init((hashid),(void *)(secret),(secret_len),mhash_get_hash_pblock(hashid));	\
		mhash(td,(data),(data_len));	\
		(hmac) = mhash_hmac_end(td);	\
	} while(0)
#define MHASH_HMAC_FREE(hmac) free((hmac))


#define PRF_SHA1	MHASH_SHA1
#define PRF_TIGER	MHASH_TIGER
#define PRF_RIPEMD	MHASH_RIPEMD160

#define PRF_MAX_HASH_SIZE	32 /* (192bits == 24 bytes) < 32 */
#define PRF_MAX_SECRET_SIZE	256
#define PRF_MAX_SEED_SIZE	256


typedef struct rand_ctx {
	int fd;
} rand_ctx;

static pthread_once_t	__rand_once_control = PTHREAD_ONCE_INIT;
static pthread_key_t	__rand_ctx_key;

void rand_init(void);
void __rand_once_init(void);
void __rand_thread_ctx_free(void *);
void rand_get(rand_ctx *,void *,int);
rand_ctx *rand_getctx(void);

void prf_expand(int, unsigned char *, int, unsigned char *, int, unsigned char *, int);



#define DH_MAX_PRIME_BIT_SIZE   4096
#define DH_MAX_PRIME_BYTE_SIZE  (DH_MAX_PRIME_BIT_SIZE/8)
#define DH_MAX_SECRET_BYTE_SIZE DH_MAX_PRIME_BYTE_SIZE


typedef struct dh_t {
	int p_size_bits;
	int p_size_bytes;

	mpz_t g; /* generator	*/
	mpz_t p; /* prime	*/
	mpz_t x; /* private key	*/
	mpz_t y; /* public key	*/
	mpz_t z; /* secret	*/
} dh_t;


void dh_init(dh_t *);
void dh_generate_parameters(dh_t *,int,int);
void dh_set_prime(dh_t *,mpz_t);
void dh_set_generator(dh_t *,mpz_t);
void dh_generate_key(dh_t *);
void dh_compute_key(dh_t *,mpz_t);
void dh_destroy(dh_t *);


#endif /* _CRYPT_H */
