/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <netinet/in.h>

#ifndef _MCRYPT_H
#define _MCRYPT_H
#include <mcrypt.h>
#endif

#include "common.h"
#include "thread.h"
#include "log.h"
#include "crypt.h"
#include "bignum.h"
#include "packet.h"
#include "node.h"


node_t *node_create(int sockd) {
	node_t *node;

	if ((node = malloc(sizeof(node_t))) == NULL) {
		ERROR("could not malloc space for a new node");
		return NULL;
	}
	bzero(node,sizeof(node_t));
	
	object_init((object_t *)node,node_free,node);
	thread_init((thread_t *)node,node_run,node);

	((thread_t *)node)->name = "node";

	dh_init(&node->dh);
	node->caller_sockd	= sockd;
	node->receiver_sockd	= 0;

	return node;
} /* end node_create */


void node_free(void *node) {

	dh_destroy(&((node_t *)node)->dh);

	thread_destroy(node);
	object_destroy(node);

	free(node);
	
	return; /* void */
} /* end node_free */


void node_server_call_back(int sockd) {
	node_t *node;
	
	if ((node = node_create(sockd)) == NULL) {
		ERROR("unable to create node");
		return; /* void */
	}

	thread_run(node);
	
	return; /* void */

} /* end node_server_call_back */


void node_run(void *__node) {
	node_t *node;	

	node = (node_t *)__node;

	if (node_link_initiate(node) == false)
		goto finish;

	if (node_caller_initiate(node) == false)
		goto finish;

	node_control_loop(node);


	finish:

	INFO("node finishing");

	close(node->caller_sockd);
	close(node->receiver_sockd);
	
	return; /* void */
	
} /* node_run */


bool node_link_initiate(node_t *node) {
	unsigned char packet[PACKET_SIZE];

	if (read_packet(node->caller_sockd,packet) != PACKET_SIZE) {
		WARN("i/o problem during link initiation");
		return false;
	}

	if (!FLAG_ISSET(packet[0],PACKET_LINK_INIT)) {
		WARN("protocol violation during link initiation: INIT flag not set");
		return false;
	}

	if ((node->hopid = packet[1]) > 4) {
		WARN("protocol voilation during link initiation: hopid of %i",packet[1]);
		return false;
	}


	if (packet[2] > PROTOCOL_MAJOR ||
	    packet[3] > PROTOCOL_MINOR ||
	    packet[4] > PROTOCOL_PATCH) {
	
		WARN("protocol version request - %i.%i.%i - not supported",packet[2],packet[3],packet[4]);
		return false;
	}

	ERROR("link established; hopid: %i",node->hopid);

	return true;

} /* end node_link_initiate */


bool node_caller_initiate(node_t *node) {
	int i, b, bytes,p_size,g_size,y_size,z_size;
	mpz_t n;
	unsigned char *buf;
	unsigned char packet[PACKET_SIZE];
	unsigned char input[NODE_CALLER_INITIATE_INPUT_SIZE]; /* one large block */
	unsigned char *cipher		= &input[0];
	unsigned char *hash		= &input[1];
	unsigned char *_p_size		= &input[2];
	unsigned char *_g_size		= &input[3];
	unsigned char *_y_size		= &input[4];
	unsigned char *prime		= &input[5];
	unsigned char *generator	= &prime[NODE_MAX_P_SIZE];
	unsigned char *pubk		= &generator[NODE_MAX_G_SIZE];
	unsigned char *seed		= &pubk[NODE_MAX_Y_SIZE];
	unsigned char secret[NODE_MAX_P_SIZE];
	unsigned char key_material[16 + 16 + 16 + 64 + 2 + 2];
	unsigned char *decrypt_key	= &key_material[0]; /* switched from the caller */
	unsigned char *encrypt_key	= &decrypt_key[16]; /* switched from the caller */
	unsigned char *cipher_iv	= &encrypt_key[16];
	unsigned char *hmac_key		= &cipher_iv[16];
	uint16_t *seq_init		= (uint16_t *)&hmac_key[64];

	bytes	= NODE_CALLER_INITIATE_INPUT_SIZE;
	buf	= input;
	
	do {
		if ((b = read_packet(node->caller_sockd,packet)) != PACKET_SIZE) {
			WARN("failed to read packet in caller initiation");		
			return false;
		}
		if (!FLAG_ISSET(packet[0],PACKET_LINK_NBND)) {
			printf("Received packet from caller:\n");
			PRINT_BYTES(stdout,packet,PACKET_SIZE);
			continue;
		}
		if (!FLAG_ISSET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_INIT)) {
			WARN("got packet with bad node flag set");
			return false;
		}
		memcpy(buf,&packet[PACKET_DATA_POS(node->hopid)],PACKET_DATA_SIZE);
		buf	+= PACKET_DATA_SIZE;
		bytes	-= PACKET_DATA_SIZE;
	} while (bytes > 0);

	p_size = (*_p_size + 1);
	g_size = (*_g_size + 1);
	y_size = (*_y_size + 1);


	if (ENABLE_DEBUG) {

		fprintf(stdout,"Recevied caller's prime:\n");
		PRINT_BYTES(stdout,prime,p_size);

		fprintf(stdout,"Recevied caller's generator:\n");
		PRINT_BYTES(stdout,generator,g_size);

		fprintf(stdout,"Received caller's pubk:\n");
		PRINT_BYTES(stdout,pubk,y_size);

	} /* DEBUG */


	/* DH secret generation */
	mpz_init(n);
	
	bignum_parse_u(n,p_size,prime);
	dh_set_prime(&node->dh,n);
	bignum_parse_u(n,g_size,generator);
	dh_set_generator(&node->dh,n);

	bignum_parse_u(n,y_size,pubk);

	dh_generate_key(&node->dh);
	dh_compute_key(&node->dh,n);
	
	z_size = bignum_format_u(node->dh.z,secret);

	if (ENABLE_DEBUG) {
		fprintf(stdout,"Node's shared secret:\n");
		PRINT_BYTES(stdout,secret,z_size);
	} /* DEBUG */


	/* re-use pubk space for our own pubk */
	bzero(pubk,NODE_MAX_Y_SIZE);
	y_size = bignum_format_u(node->dh.y,pubk);

	if (ENABLE_DEBUG) {
		printf("Node's pubk:\n");
		PRINT_BYTES(stdout,pubk,y_size);
	} /* DEBUG */

	mpz_clear(n);
	/* end DH secret generation */


	/* setup hash */
	switch (*hash) {
		case NODE_HASH_TIGER:
			node->hash_type = TIGER_192;
			break;
		case NODE_HASH_SHA1:
			node->hash_type = SHA1_160;
			break;
		default:
			WARN("bad hash requested in caller initiation: %i",*hash);
			return false;
	} /* end hash setup */


	/* create keying/sequencing material */
	prf_expand(node->hash_type,key_material,sizeof(key_material),secret,z_size,seed,NODE_SEED_SIZE);


	/* setup ciphers */
	/*switch(*cipher) {
		case NODE_CIPHER_AES:
			if ((node->encrypt_td = mcrypt_module_open(MCRYPT_RIJNDAEL_128,NULL,MCRYPT_CBC,NULL)) == MCRYPT_FAILED) {
				ERROR("failed to open mcrypt cipher %s",MCRYPT_RIJNDAEL_128);
				return false;
			}
			if ((node->decrypt_td = mcrypt_module_open(MCRYPT_RIJNDAEL_128,NULL,MCRYPT_CBC,NULL)) == MCRYPT_FAILED) {
				ERROR("failed to open mcrypt cipher %s",MCRYPT_RIJNDAEL_128);
				return false;
			}
			break;
		case NODE_CIPHER_TWOFISH:
			if ((node->encrypt_td = mcrypt_module_open(MCRYPT_TWOFISH,NULL,MCRYPT_CBC,NULL)) == MCRYPT_FAILED) {
				ERROR("failed to open mcrypt cipher %s",MCRYPT_TWOFISH);
				return false;
			}
			if ((node->decrypt_td = mcrypt_module_open(MCRYPT_RIJNDAEL_128,NULL,MCRYPT_CBC,NULL)) == MCRYPT_FAILED) {
				ERROR("failed to open mcrypt cipher %s",MCRYPT_RIJNDAEL_128);
				return false;
			}
			break;
		default:
			WARN("bad cipher requested in caller initiation: %i",*cipher);
			return false;
	}
	mcrypt_generic_init(node->encrypt_td,encrypt_key,16,cipher_iv);
	mcrypt_generic_init(node->decrypt_td,decrypt_key,16,cipher_iv);*/
	
	cipher_init(&node->encrypt_cipher,*cipher,encrypt_key,cipher_iv);
	cipher_init(&node->decrypt_cipher,*cipher,decrypt_key,cipher_iv);
	
	/* end cipher setup */


	memcpy(node->hmac_key,hmac_key,64);
	node->caller_seq = ntohs(*seq_init);


	if (ENABLE_DEBUG) {
		printf("Node's encrypt key:\n");
		PRINT_BYTES(stdout,encrypt_key,16);

		printf("Node's decrypt key:\n");
		PRINT_BYTES(stdout,decrypt_key,16);

		printf("Node's hmac key:\n");
		PRINT_BYTES(stdout,node->hmac_key,64);
	}


	/*	
	 * everything is setup on our side.
	 * time to respond. we can reuse all of our buffers
	 *
	 */

	input[0] = (y_size - 1); /* shift (1,256) -> (0,255) */
	memcpy(&input[1],pubk,NODE_MAX_Y_SIZE);

	bytes	= 1 + NODE_MAX_Y_SIZE;
	buf	= input;

	FLAG_CLR(packet[0]);
	FLAG_SET(packet[0],PACKET_LINK_NBND);

	
	FLAG_CLR(packet[PACKET_FLAG_POS(node->hopid)]);
	FLAG_SET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_INIT);

	do {
		memcpy(&packet[PACKET_DATA_POS(node->hopid)],buf,MIN(PACKET_DATA_SIZE,bytes));
		HMAC(node->hash_type,&packet[PACKET_MAC_POS(node->hopid)],node->hmac_key,sizeof(node->hmac_key),&packet[PACKET_HASH_BEG(node->hopid)],PACKET_HASH_LEN(node->hopid));

		if (write_packet(node->caller_sockd,packet) < PACKET_SIZE) {
			WARN("packet write failed while completing initialization");
			return false;
		}
		
		bytes	-= PACKET_DATA_SIZE;
		buf	+= PACKET_DATA_SIZE;
		
	} while (bytes > 0);

	ERROR("caller initiated; hopid: %i",node->hopid);

	return true;

} /* end node_caller_initiate */


void node_control_loop(node_t *node) {
	unsigned char packet[PACKET_SIZE];
	unsigned char hmac[64];
	rand_ctx *ctx;
	struct timeval tv;
	int i,max_fd = 0;
	fd_set fdset;

	DEBUG("entering node control loop");

	ctx = rand_getctx();

	for (;;) {
		FD_ZERO(&fdset);
		FD_SET(node->caller_sockd,&fdset);
		if (node->receiver_sockd > 0) {
			FD_SET(node->receiver_sockd,&fdset);
		}

		max_fd = (MAX(node->caller_sockd,node->receiver_sockd) + 1);

		tv.tv_sec	= 1;
		tv.tv_usec	= 500;

		if ((i = select(max_fd,&fdset,NULL,NULL,&tv)) < 0) {
			WARN("select returned an error: %s",strerror(errno));
			goto finish;
		}

		DEBUG("select returned w/ %i %s",i,(i == 1)?"descriptor":"descriptors");
		
		if (FD_ISSET(node->caller_sockd,&fdset)) {

			DEBUG("reading data from caller's side");

			if (read_secure_packet(node,packet) != PACKET_SIZE) {
				WARN("unable to read full packet");
				goto finish;
			}

			if (packet[0] == PACKET_LINK_NULL) {
				DEBUG("got NULL packet from link");
				continue;
			}
			
			if (packet[0] == PACKET_LINK_KILL) {
				INFO("got kill request on link");
				goto finish;
			}

			if (FLAG_ISSET(packet[0],PACKET_LINK_NBND)) {
				
				if (FLAG_ISSET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_FWRD)) {
					DEBUG("forwarding packet...");
					if (node->receiver_sockd > 0) {
						rand_get(ctx,&packet[PACKET_HEAD_BEG(node->hopid)],PACKET_HEAD_LEN(node->hopid));
						if (write_packet(node->receiver_sockd,packet) != PACKET_SIZE) {
							WARN("unable to forward packet");
							goto finish;
						}
						DEBUG("forwarded packet for caller");
					} else {
						WARN("protocol violation: request to forward packet to non-existant node");
						goto finish;
					
					}
				} else if (FLAG_ISSET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_CNCT)) {
					if (node->receiver_sockd > 0) {
						WARN("protocol violation in packet: cannot connect to new node; already connected");
						goto finish;
					} else {
						if (node_do_connect(node,packet) == false) {
							WARN("node connect failed");
							goto finish;
						}
						DEBUG("succeeded in proxying for caller");
					}
				} else if (FLAG_ISSET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_KILL)) {
					DEBUG ("got node kill request");
					goto finish;
				} else {
					WARN("protocol violation: unexpected flag set in packet from caller");
					goto finish;
				}
	
			} else {
				WARN("protocol violation in packet link header");
				goto finish;
			}

		}

		if (node->receiver_sockd > 0 && FD_ISSET(node->receiver_sockd,&fdset)) {

			DEBUG("reading data from receiver's side");
		
			if (read_packet(node->receiver_sockd,packet) != PACKET_SIZE) {
				WARN("unable to read packet; closing receiving side");
				close(node->receiver_sockd);
				node->receiver_sockd = 0;
				continue;
			}
			
			if (write_secure_packet(node,packet) != PACKET_SIZE) {
				WARN("unable to write packet");
				goto finish;
			}
		}
		
	}

	finish:

	DEBUG("node quiting control loop");

	return; /* void */

} /* end node_control_loop */


bool node_do_connect(node_t *node,unsigned char *packet) {
	bool rval = false;
	struct sockaddr_in sin;
	struct hostent *hostent;
	unsigned char request[1 + 256 + 2];	/* flags + address + port */
	unsigned char response[1 + 16];		/* flags + ipaddress */
	char host[257];
	unsigned char *buf;
	int bytes;

	bytes	= sizeof(request);
	buf	= request;
	memcpy(buf,&packet[PACKET_DATA_POS(node->hopid)],PACKET_DATA_SIZE);
	bytes	-= PACKET_DATA_SIZE;
	buf	+= PACKET_DATA_SIZE;
	

	do {
		if (read_secure_packet(node,packet) != PACKET_SIZE) {
			WARN("unable to read packet from caller during node connection establishment");
			response[0] = 0;
			goto finish;
		}


		if (packet[0] == PACKET_LINK_NULL) {
			continue;
		} else if (packet[0] == PACKET_LINK_KILL) {
			WARN("got link kill request");
			node->state = NODE_STATE_DEAD;
			return false;
		} else if (FLAG_ISSET(packet[0],PACKET_LINK_NBND) && FLAG_ISSET(packet[PACKET_FLAG_POS(node->hopid)],PACKET_NODE_CNCT)) {
		
			memcpy(buf,&packet[PACKET_DATA_POS(node->hopid)],MIN(bytes,PACKET_DATA_SIZE));
			bytes	-= PACKET_DATA_SIZE;
			buf	+= PACKET_DATA_SIZE;
		
		} else {
			WARN("protocol violation in link header or node header");
			node->state = NODE_STATE_DEAD;
			return false;
		}
	
	} while (bytes > 0);

	if (FLAG_ISSET(request[0],PACKET_CNCT_REQ_IPV4)) {
		sin.sin_family	= AF_INET;
		memcpy(&sin.sin_addr,&request[1],4);
		memcpy(&response[1],&request[1],4);
		sin.sin_port	= *(uint16_t *)&request[257];
		FLAG_SET(response[0],PACKET_CNCT_RSP_IPV4);
	} else if (FLAG_ISSET(request[0],PACKET_CNCT_REQ_IPV6)) {
		WARN("request to connect to IPv6 address: cannot use IPv6 yet");
		response[0] = 0;
		goto finish;
	} else if (FLAG_ISSET(request[0],PACKET_CNCT_REQ_NAME)) {
		bzero(host,257);
		memcpy(host,&request[1],256);
		sin.sin_port = *(uint16_t *)&request[257];

		if ((hostent = gethostbyname(host)) == NULL) {
			WARN("failed to resolve hostname in request");
			response[0] = 0;
			goto finish;
		}

		if (hostent->h_addrtype == AF_INET6) {
			WARN("hostname resolved to an IPv6 address: cannot use IPv6 yet");
			response[0] = 0;
			goto finish;
		}

		if (hostent->h_addr_list[0] != NULL) {
			DEBUG("length of h_addr's: %i",hostent->h_length);
			memcpy(&sin.sin_addr,hostent->h_addr_list[0],4);
			memcpy(&response[1],hostent->h_addr_list[0],4);
			FLAG_SET(response[0],PACKET_CNCT_RSP_IPV4);
		} else {
			response[0] = 0;
			goto finish;
		}

	} else {
		WARN("protocol violation in connection request");
		response[0] = 0;
		goto finish;
	}


	if (node_link_connect(node,&sin) == false) {
		WARN("unable to setup link to request node");
		response[0] = 0;
		goto finish;
	}

	rval = true; /* everything is fine so far */

	finish:
	
	bytes	= sizeof(response);
	buf	= response;

	do {
		memcpy(&packet[PACKET_DATA_POS(node->hopid)],buf,MIN(PACKET_DATA_SIZE,bytes));
		packet[PACKET_PLEN_POS(node->hopid)] = MIN(PACKET_DATA_SIZE,bytes);
		bytes	-= PACKET_DATA_SIZE;
		buf	+= PACKET_DATA_SIZE;

		if (write_secure_packet(node,packet) != PACKET_SIZE) {
			WARN("unable to write packet during connection establishment");
			return false;
		}

	} while (bytes > 0);	

	return rval;

} /* end node_do_connect */


bool node_link_connect(node_t *node,struct sockaddr_in *sin) {
	unsigned char packet[PACKET_SIZE];

	if ((node->receiver_sockd = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) {
		ERROR("could not create a socket: %s",strerror(errno));
		return false;
	}

	if (connect(node->receiver_sockd,(struct sockaddr *)sin,(socklen_t)sizeof(struct sockaddr_in)) != 0) {
		ERROR("could not connect to requested node: %s",strerror(errno));
		return false;
	}

	bzero(packet,PACKET_SIZE);
	FLAG_SET(packet[0],PACKET_LINK_INIT);
	packet[1]	= (node->hopid + 1);	
	packet[2]	= PROTOCOL_MAJOR;
	packet[3]	= PROTOCOL_MINOR;
	packet[4]	= PROTOCOL_PATCH;

	if (write_packet(node->receiver_sockd,packet) != PACKET_SIZE) {
		WARN("unable to write packet during link initiation");
		return false;
	}

	DEBUG("setup link to node for caller");

	return true;

} /* end node_link_connect */


int read_secure_packet(node_t *node,unsigned char *packet) {
	unsigned char hmac[64];

	if (read_packet(node->caller_sockd,packet) != PACKET_SIZE)
		return 0;

	//mdecrypt_generic(node->decrypt_td,&packet[PACKET_ENC_BEG(node->hopid)],PACKET_ENC_LEN(node->hopid));
	cipher_decrypt(&node->decrypt_cipher,&packet[PACKET_ENC_BEG(node->hopid)],PACKET_ENC_LEN(node->hopid));
	HMAC(node->hash_type,hmac,node->hmac_key,sizeof(node->hmac_key),&packet[PACKET_HASH_BEG(node->hopid)],PACKET_HASH_LEN(node->hopid));
	if (memcmp(hmac,&packet[PACKET_MAC_POS(node->hopid)],10) != 0) {
		WARN("bad MAC in packet");
		/* continue anyhow for now */
	}
				
	/* insert check sequence here */

	return PACKET_SIZE;
} /* end read_secure_packet */


int write_secure_packet(node_t *node,unsigned char *packet) {
	unsigned char hmac[64];

	/* insert seq stuff here */

	HMAC(node->hash_type,hmac,node->hmac_key,sizeof(node->hmac_key),&packet[PACKET_HASH_BEG(node->hopid)],PACKET_HASH_LEN(node->hopid));
	memcpy(&packet[PACKET_MAC_POS(node->hopid)],hmac,10);
	//mcrypt_generic(node->encrypt_td,&packet[PACKET_ENC_BEG(node->hopid)],PACKET_ENC_LEN(node->hopid));
	cipher_encrypt(&node->encrypt_cipher,&packet[PACKET_ENC_BEG(node->hopid)],PACKET_ENC_LEN(node->hopid));

	return write_packet(node->caller_sockd,packet);

} /* end write_secure_packet */


int read_packet(int sockd,void *packet) {
	int b, bytes;
	unsigned char *buf;

 	buf	= (unsigned char *)packet;
	bytes	= PACKET_SIZE;

	do {
		if ((b = read(sockd,buf,bytes)) < 1) {
			if (b == 0) {
				WARN("premature EOF on socket");
			} else {
				ERROR("socket read failed: %s",strerror(errno));
			}
			break;
		}
		bytes	-= b;
		buf	+= b;
	} while (bytes > 0);

	return PACKET_SIZE - bytes;

} /* end read_packet */


int write_packet(int sockd,void *packet) {
	int b, bytes;
	unsigned char *buf;

	buf	=  (unsigned char *)packet;
	bytes	= PACKET_SIZE;
	
	do {
		if ((b = write(sockd,buf,bytes)) < 0) {
			WARN("socket write failed: %s",strerror(errno));
			break;
		}
		bytes	-= b;
		buf	+= b;
	} while (bytes > 0);

	return PACKET_SIZE - bytes;
	
} /* end write_packet */

