/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include "common.h"
#include "log.h"
#include "object.h"


void object_init(void *arg,void (*destructor)(void *),void *destructor_arg) {
	object_t *obj = (object_t *)arg;

	/* initialize members */
	obj->__free		= destructor;
	obj->__free_arg		= (destructor_arg == NULL)? obj : destructor_arg;

	obj->lock	= object_lock;
	obj->unlock	= object_unlock;

	obj->retain	= object_retain;
	obj->release	= object_release;
	
	obj->wait	= object_wait;
	obj->timedwait	= object_timedwait;
	obj->signal	= object_signal;
	obj->broadcast	= object_broadcast;



	/* initialize mutexes */
	pthread_mutexattr_init(&obj->__sync_mutexattr);
	pthread_mutexattr_init(&obj->__gc_mutexattr);
	pthread_mutexattr_init(&obj->__cond_mutexattr);

	pthread_condattr_init(&obj->__cond_attr);

	/* these are causing seg faults somehow, somewhere */
	/*pthread_mutexattr_settype(&obj->__sync_mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutexattr_settype(&obj->__gc_mutexattr,PTHREAD_MUTEX_DEFAULT);
	pthread_mutexattr_settype(&obj->__cond_mutexattr,PTHREAD_MUTEX_DEFAULT);*/

	pthread_mutex_init(&obj->__sync_mutex,&obj->__sync_mutexattr);
	pthread_mutex_init(&obj->__gc_mutex,&obj->__gc_mutexattr);
	pthread_mutex_init(&obj->__cond_mutex,&obj->__cond_mutexattr);

	pthread_cond_init(&obj->__cond_var,&obj->__cond_attr);

	/* automatically retain once */
	obj->__gc_ref_count = 1;

	return; /* void */

} /* end object_init */


void object_destroy(void *obj) {

	/* destroy mutexes */
	pthread_mutex_destroy(&((object_t *)obj)->__sync_mutex);
	pthread_mutex_destroy(&((object_t *)obj)->__gc_mutex);
	pthread_mutex_destroy(&((object_t *)obj)->__cond_mutex);
	
	pthread_mutexattr_destroy(&((object_t *)obj)->__sync_mutexattr);
	pthread_mutexattr_destroy(&((object_t *)obj)->__gc_mutexattr);
	pthread_mutexattr_destroy(&((object_t *)obj)->__cond_mutexattr);

	return; /* void */

} /* end object_destroy */


object_t *object_create() {
	object_t *obj;

	if ((obj = malloc(sizeof(object_t))) == NULL)
		return NULL;

	object_init(obj,__object_free,obj);
	
	return obj;

} /* end object_create */


void object_free(object_t *obj) {

	object_destroy(obj);
	free(obj);
	
	return; /* void */

} /* end object_free */


void __object_free(void *obj) {
	object_free((object_t *)obj);
} /* end __object_free */


/* NOTE: if retain is called after this object has been
   destroyed (or a release is called that will destroy it,
   then that is a programming error and not a bug here.
   a thread should retain an object before the calling
   thread that passed it continues.
   NOTE: do retain and release need their own lock? */
void object_retain(void *obj) {

	pthread_mutex_lock(&((object_t *)obj)->__gc_mutex);

	((object_t *)obj)->__gc_ref_count++;	

	pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex);
	
	return; /* void */

} /* end object_retain */
 

/* NOTE: race condition? will pthread_mutex_lock return properly
   if the mutex is destroyed while another is IN pthread_mutex_lock */
void object_release(void *obj) {

	if (pthread_mutex_lock(&((object_t *)obj)->__gc_mutex) == 0) {
		((object_t *)obj)->__gc_ref_count--;


		pthread_mutex_unlock(&((object_t *)obj)->__gc_mutex);
	}

	if (((object_t *)obj)->__gc_ref_count == 0) {
		if (((object_t *)obj)->__free != NULL)
			((object_t *)obj)->__free(((object_t *)obj)->__free_arg);
		return; /* void */
	}

	return; /* void */

} /* object_release */


void object_lock(void *obj) {

	DEBUG("locking object");

	pthread_mutex_lock(&((object_t *)obj)->__sync_mutex);
	
	DEBUG("locked object");
	
	return; /* void */

} /* end object_lock */


void object_unlock(void *obj) {

	DEBUG("unlocking object");

	pthread_mutex_unlock(&((object_t *)obj)->__sync_mutex);

	DEBUG("unlocked object");

	return; /* void */

} /* end object_unlock */


void object_wait(void *obj) {

	pthread_mutex_lock(&((object_t *)obj)->__cond_mutex);

	pthread_cond_wait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__cond_mutex);

	pthread_mutex_unlock(&((object_t *)obj)->__cond_mutex);

	return; /* void */

} /* end object_wait */


void object_timedwait(void *obj,const struct timespec *tp) {

	pthread_mutex_lock(&((object_t *)obj)->__cond_mutex);

	pthread_cond_timedwait(&((object_t *)obj)->__cond_var,&((object_t *)obj)->__cond_mutex,tp);

	pthread_mutex_unlock(&((object_t *)obj)->__cond_mutex);

	return; /* void */

} /* end object_timedwait */


void object_signal(void *obj) {

	pthread_cond_signal(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_signal */


void object_broadcast(void *obj) {

	pthread_cond_broadcast(&((object_t *)obj)->__cond_var);

	return; /* void */

} /* end object_broadcast */

