/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>

#include "common.h"
#include "log.h"
#include "thread.h"
#include "list.h"


List list_create(size_t size) {
	List list;
	void list_destroy(List);
	void list_bind(List);
	void list_release(List);
	void list_lock(List);
	void list_unlock(List);
	size_t list_size(List);
	size_t list_count(List);
	bool list_grow(List,size_t);
	bool list_shrink(List,size_t);
	void *list_pop(List);
	void *list_shift(List);
	void *list_push(List,void *);
	void *list_unshift(List,void *);
	
	if ((list = (List)malloc(sizeof(list_t))) == NULL) {
		WARN("unable to create new list");
		return NULL;
	}
	bzero(list,sizeof(list_t));

	if ((list->object = object_create(&list_destructor,list)) == NULL) {
		WARN("unable to get new object for list");
		list_destroy(list);
		return NULL;
	}

	DEBUG("got object");

	list->_min = list->_size = size;
	list->_count = 0;

	if ((list->_list = (void **)malloc(sizeof(int)*size)) == NULL) {
		WARN("unable to malloc space for list items");
		list_destroy(list);
		return NULL;
	}

	list->bind    = &list_bind;
	list->release = &list_release;
	list->lock    = &list_lock;
	list->unlock  = &list_unlock;

	list->size    = &list_size;
	list->count   = &list_count;
	list->grow    = &list_grow;
	list->shrink  = &list_shrink;

	list->push    = &list_push;
	list->pop     = &list_pop;
	list->unshift = &list_unshift;
	list->shift   = &list_shift;
	
	return list;
}

void list_destroy(List list) {
	if (list->object != NULL)
		object_destroy(list->object);
	free(list);
}

void list_destructor(void *arg) {
	list_destroy((List)arg);
}

void list_bind(List this) {
	this->object->bind(this->object);
}

void list_release(List this) {
	this->object->release(this->object);
}

size_t list_size(List this) {
	return this->_size;
}

size_t list_count(List this) {
	return this->_count;
}

bool list_grow(List this,size_t amt) {
	void **tmp;

	this->_size +=amt;
	if ((tmp = realloc((void *)this->_list,sizeof(int)*this->_size)) == NULL) {
		WARN("unable to realloc space for list: %s",strerror(errno));
		this->_size -= amt;
		return false;
	}
	this->_list = tmp;

	return true;
}

bool list_shrink(List this,size_t amt) {
	void *tmp;

	if ((tmp = realloc((void *)this->_list,sizeof(int) * MAX((this->_size-=amt),MAX(this->_count,this->_min)))) == NULL) {
		WARN("unable to realloc space for list");
		this->_size += amt;
		return false;
	}
	this->_list = (void **)tmp;
	return true;
}

void list_lock(List this) {
	this->object->lock(this->object);
}

void list_unlock(List this) {
	this->object->unlock(this->object);
}

void *list_push(List this,void *item) {
	if (this->_size == this->_count) {
		unless (this->grow(this,1))
			return NULL;
	}

	this->_list += this->_count;
	*this->_list = item;
	this->_list -= this->_count++;

	return item;
}

void *list_pop(List this) {
	void *item;
	this->_list += --this->_count;
	item = *this->_list;
	this->_list -= this->_count;
	return item;
}

void *list_unshift(List this,void *item) {
	void **tmp;

	if (this->_size == this->_count) {
		unless (this->grow(this,1))
			return NULL;
	}

	tmp = this->_list;
	
	tmp++;

	memmove(tmp,this->_list,(sizeof(int) * this->_count));
	this->_count++; 
	*this->_list = item;

	return item;
}

void *list_shift(List this) {
	void *item;
	void **tmp;

	item = *this->_list;

	tmp = this->_list;
	tmp++;
	this->_count--;
	memmove(this->_list,tmp,(sizeof(int) * this->_count));


	return item;
}
