/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */

#include <config.h>

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <openssl/bn.h>
#include <openssl/dh.h>

#include "common.h"
#include "thread.h"
#include "log.h"
#include "packet.h"
#include "node.h"
#include "link.h"
#include "prng.h"
#include "crypt.h"

Node node_create(int sockfd) {
	Node node;
	void node_bind(Node);
	void node_release(Node);
	void node_lock(Node);
	void node_unlock(Node);

	bool node_start(Node);
	bool node_is_alive(Node);

	void node_destructor(void *);
	void *node_thread_entry(void *);

	void node_set_hop(Node,int);
	void node_set_inner(Node,int);


	if ((node = (Node)malloc(sizeof(node_t))) == NULL) {
		ERROR("could not malloc space for a new node");
		return NULL;
	}
	bzero(node,sizeof(node_t));

	if ((node->inner = (Link)malloc(sizeof(link_t))) == NULL) {
		ERROR("could not malloc space for a new link");
		return NULL;
	}
	bzero(node->inner,sizeof(link_t));

	if ((node->outer = (Link)malloc(sizeof(link_t))) == NULL) {
		ERROR("could not malloc space for a new link");
		return NULL;
	}
	bzero(node->outer,sizeof(link_t));

	if ((node->object = object_create(&node_destructor,node)) == NULL) {
		ERROR("could not get new object");
		node_destroy(node);
		return NULL;
	}

	if ((node->thread = thread_create(&node_thread_entry,node)) == NULL) {
		ERROR("could not get new thread");
		node_destroy(node);
		return NULL;
	}

	if ((node->inner->dh = node->dh = DH_new()) == NULL) {
		ERROR("could not get a new DH object");
		node_destroy(node);
		return NULL;
	}

	if ((node->inner->td = node->td = mcrypt_module_open(MCRYPT_RIJNDAEL_128,NULL,MCRYPT_CBC,NULL)) == NULL) {
		ERROR("could not get new Mcrypt object");
		node_destroy(node);
		return NULL;
	}

	node->bind      = &node_bind;
	node->release   = &node_release;
	node->lock      = &node_lock;
	node->unlock    = &node_unlock;

	node->start     = &node_start;
	node->isAlive   = &node_is_alive;

	node->setHop = &node_set_hop;

	node->inner->readp  = &link_readp;
	node->inner->writep = &link_writep;
	node->outer->readp  = &link_readp;
	node->outer->writep = &link_writep;
	node->inner->in_seq = 0;
	node->inner->out_seq = 0;
	node->outer->in_seq = 0;
	node->outer->out_seq = 0;

	node->inner->read  = &link_read;
	node->inner->write = &link_write;
	node->outer->read  = &link_read;
	node->outer->write = &link_write;

	node->inner->sockfd = sockfd;
	node->outer->sockfd = 0;

	return node;
}

void node_destroy(Node this) {
	if (this == NULL)
		return;

	thread_destroy(this->thread);
	object_destroy(this->object);

	DH_free(this->dh);
	mcrypt_module_close(this->td);

	free(this->inner);
	free(this->outer);
	free(this);
}

void node_destructor(void *arg) {
	node_destroy((Node)arg);
}

void node_bind(Node this) {
	this->object->bind(this->object);
}

void node_release(Node this) {
	this->object->release(this->object);
}

void node_lock(Node this) {
	this->object->lock(this->object);
}

void node_unlock(Node this) {
	this->object->unlock(this->object);
}

bool node_start(Node this) {
	if (this->thread->start(this->thread)) {
		this->bind(this);
		return true;
	} else {
		ERROR("could not start node");
		return false;
	}
}

bool node_is_alive(Node this) {
	return this->thread->isAlive(this->thread);
}

void *node_thread_entry(void *arg) {
	node_run((Node)arg);
	((Node)arg)->thread->setAlive(((Node)arg)->thread,false);
	((Node)arg)->release((Node)arg);
	return NULL;
}

void node_run(Node this) {
	struct timeval tv;
	int max_fd = 0;
	int b,c,i;
	fd_set fdset;
	unsigned char type;

	DEBUG("\n====================================================");
	DEBUG("node started....");

	if (node_link_init(this) == false) {
		WARN("failed link initiation");
		goto finish;
	}

	if (node_node_init(this) == false) {
		WARN("failed node initiation");
		goto finish;
	}

	sleep(1);
	if (this->inner->readp(this->inner)) {
		DEBUG("got packet: %s",hex(this->inner->recvbuf,129));
	} else {
		DEBUG("read failed");
	}


	sleep(1);
	if (this->inner->readp(this->inner)) {
		DEBUG("got packet: %s",hex(this->inner->recvbuf,129));
	} else {
		DEBUG("read failed");
	}

	goto finish;

	max_fd = this->inner->sockfd;

	DEBUG("entering node i/o loop");
	while(1) {
		FD_ZERO(&fdset);

		FD_SET(this->inner->sockfd,&fdset);
		
		tv.tv_sec  = 0;
		tv.tv_usec = 300;

		if (select(max_fd,&fdset,NULL,&fdset,&tv) < 0) {
			WARN("select returned an error: %s",strerror(errno));
			goto finish;
		}

		if (FD_ISSET(this->inner->sockfd,&fdset)) {
			if (this->inner->readp(this->inner)) {
				
			} else {
				WARN("select return but couldn't read a new packet");
			}
		} else {
			WARN("what did select return?");
			goto finish;
		}
	}


	finish:

	DEBUG("finishing....");
	close(this->inner->sockfd);

	return;
}

void node_set_hop(Node this,int hopid) {
	this->hopid = this->outer->hopid = this->inner->hopid = hopid;
}


bool node_link_init(Node this) {
	unsigned char buf[PACKET_SIZE];

	DEBUG("starting link initialization");
	this->inner->lflag = PACKET_LINK_INIT;

	if (this->inner->readp(this->inner) && this->inner->recvbuf[1] < 5) {
		this->setHop(this,(int) buf[0]);
	} else {
		return false;
	}
}

bool node_node_init(Node this) {
	unsigned char buf[MAX(1024,4+NODE_MAX_P_SIZE+NODE_MAX_G_SIZE+NODE_MAX_PUBK_SIZE+NODE_MAX_SEED_SIZE)];
	unsigned char *p    = NULL;
	unsigned char *g    = NULL;
	unsigned char *pubk = NULL;
	unsigned char *data = NULL;
	unsigned char seed[NODE_MAX_SEED_SIZE];
	BIGNUM *bn_p;
	BIGNUM *bn_pubk;
	int p_size,g_size,pubk_size,seed_size,secret_size;
	int in = 0,pos = 0,i;

	DEBUG("starting node initialization");

	data = &buf[0];

	FLAG_CLR(this->inner->lflag);
	FLAG_SET(this->inner->lflag,PACKET_LINK_NBND);

	FLAG_CLR(this->inner->nflag);
	FLAG_SET(this->inner->nflag,PACKET_NODE_INIT);

	if (this->inner->read(this->inner,data,4+(128*4)) == false) {
		WARN("bad packet during node initialization");
		return false;
	}

	unless (0 < (p_size = (int) *data) < 128) {
		WARN("p parameter too large: %i",p_size);
		return false;
	}
	p_size++;
	data++;
	pos++;

	unless (0 < (g_size = (int) *data) < 128) {
		WARN("g parameter too large");
		return false;
	}
	g_size++;
	data++;
	pos++;

	unless (0 < (pubk_size = (int) *data) < 128) {
		WARN("pub key parameter too large");
		return false;
	}
	pubk_size++;
	data++;
	pos++;

	unless (0 < (seed_size = (int) *data) < 128) {
		WARN("seed parameter too large");
		return false;
	}
	seed_size++;
	data++;
	pos++;

	p = data;
	data += 128;

	g = data;
	data += 128;

	pubk = data;
	data += 128;

	memcpy(seed,data,seed_size);

	if ((this->inner->dh->p = BN_bin2bn(p,p_size,NULL)) == NULL) {
		ERROR("could not get new BIGNUM");
		return false;
	}

	if ((this->inner->dh->g = BN_bin2bn(g,g_size,NULL)) == NULL) {
		ERROR("could not get new BIGNUM");
		return false;
	}

	if ((bn_pubk = BN_bin2bn(pubk,pubk_size,NULL)) == NULL) {
		ERROR("could not get a new BIGNUM");
		return false;
	}


	/* DH check? */

	if (!DH_generate_key(this->inner->dh)) {
		WARN("DH_generate_key failed");
		BN_free(bn_pubk);
		return false;
	}

	buf[0] = (unsigned char)(BN_num_bytes(this->inner->dh->pub_key)-1);

	BN_bn2bin(this->inner->dh->pub_key,&buf[1]);

	if (this->inner->write(this->inner,buf,129) == false) {
		WARN("failed to write data");
		BN_free(bn_pubk);
		return false;
	}

	if ((secret_size = DH_compute_key(buf,bn_pubk,this->inner->dh)) < 0) {
		ERROR("DH_compute_key failed");
		BN_free(bn_pubk);
		return false;
	}

	prf_expand(buf,512,buf,secret_size,seed,seed_size);
	data = buf;

	memcpy(this->cipher_key,data,16);
	memcpy(this->inner->cipher_key,data,16);
	data += 16;

	memcpy(this->hmac_key,data,16);
	memcpy(this->inner->hmac_key,data,16);
	data += 16;

	mcrypt_generic_init(this->td,this->cipher_key,16,data);
	data += 16;

	this->inner->out_seq = this->inner->in_seq = ntohs(*(unsigned short int *)data);

	this->inner->encrypted = true;

	return true;	
}
