/* ==========================================================================
 * AnonNet - Anonymous Network Project
 * --------------------------------------------------------------------------
 * Copyright (C) 2001  William Ahern
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * ==========================================================================
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <netinet/in.h>

#include <sys/types.h>
#include <sys/socket.h>

#include "common.h"
#include "thread.h"
#include "log.h"
#include "conf.h"
#include "node.h"
#include "server.h"


Server server_create() {
	Server server;
	bool server_start(Server);
	bool server_is_alive(Server);
	void server_bind(Server);
	void server_release(Server);
	void server_lock(Server);
	void server_unlock(Server);
	void server_destructor(void *);

	if ((server = (Server)malloc(sizeof(server_t))) == NULL) {
		ERROR("could not malloc space for a new server");
		return NULL;	
	}
	memset(server,0,sizeof(server_t));

	if ((server->object = object_create(&server_destructor,server)) == NULL) {
		ERROR("could not get new object");
		server_destroy(server);
	}

	if ((server->thread = thread_create(&server_thread_entry,(void *)server)) == NULL) {
		ERROR("could get a new thread");
		server_destroy(server);
		return NULL;
	}

	server->start   = &server_start;
	server->isAlive = &server_is_alive;
	server->bind    = &server_bind;
	server->release = &server_release;
	server->lock    = &server_lock;
	server->unlock  = &server_unlock;
		
	return server;
}

void server_destroy(Server this) {
	if (this->thread != NULL)
		this->thread->release(this->thread);
	object_destroy(this->object);
	free(this);
}

void server_destructor(void *arg) {
	server_destroy((Server)arg);
}

void server_bind(Server this) {
	this->object->bind(this->object);
}

void server_release(Server this) {
	this->object->release(this->object);
}

void server_lock(Server this) {
	this->object->lock(this->object);
}

void server_unlock(Server this) {
	this->object->unlock(this->object);
}

void *server_thread_entry(void *arg) {
	Server server = (Server)arg;
	server_run(server);
	server->thread->setAlive(server->thread,false);
	server->release(server);
	return NULL;
}

void server_run(Server server) {
	extern Conf conf;
	int s, sd;
	struct sockaddr_in sin;
	const socklen_t size = (socklen_t)sizeof(struct sockaddr_in);
	short port;
	Node node;
	
	port = conf->port;

	memset(&sin, 0, sizeof(size));
	sin.sin_family      = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port        = htons(port);


	if ((s = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0) { /* IP or TCP? */
		ERROR("could not create socket: %s",strerror(errno));
		goto server_end;
	}

	if (bind(s, (struct sockaddr *)&sin,size) < 0) {
		ERROR("could not bind to socket: %s",strerror(errno));
		goto server_end;
	}

	while (1) {
	
		if (listen(s,SOMAXCONN) < 0) {
			ERROR("could not listen to socket: %s",strerror(errno));
			goto server_end;
		}

		DEBUG("waiting for connection");

		if ((sd = accept(s,NULL,0)) < 0 ) {
			ERROR("could not accept connection on socket: %s",strerror(errno));
			goto server_end;
		}

		DEBUG("accepting connection");

		DEBUG("creating new node");

		if ((node = node_create(sd)) == NULL) {
			ERROR("could not get a new node");
			continue;
		}

		unless (node->start(node)) {
			ERROR("node would not start");
		}
		
	}

	server_end:
	return;
}

bool server_start(Server server) {

	if (server->thread->start(server->thread)) {
		return true;
	} else {
		ERROR("could not start server thread");
		return false;
	}
}

bool server_is_alive(Server server) {
	return server->thread->isAlive(server->thread);
}


